# -*- coding: utf-8 -*-
"""
Created on Sun Apr 5 22:40:52 2020
@author: S.Primakov
s.primakov@maastrichtuniversity.nl

"""
import os,re
import pydicom
from warnings import warn
from collections import OrderedDict
from tqdm import tqdm


is_in_list = lambda names_list, name: any([True for v in names_list if re.search(v, name.lower())])

class DataSet:
    '''This class creates imaging dataset objects.'''

    def __init__(self, data_path: str = None,
                 data_type: str = 'dcm', mask_names: list = [],
                 image_only: bool = False, multi_rts_per_pat: bool = False,
                 messy_structure: bool = False,
                 image_names: list=['image', 'volume', 'img', 'vol']):
        """Initialise a dataset object.

        Arguments:
            data_path: Path to DICOM/NRRD root directory (structure inside patient’s folder doesn’t matter).
            data_type: Original data format, 'dcm' is default, can be 'nrrd'.
            mask_names: List of names for NRRD files containing binary mask, default is ‘mask’.
            image_only: If your dataset has only images, whithout Rtstructures, you need to set it to True, otherwise default value is False.
            multi_rts_per_pat: If you have multiple Rtstructures in the patient folder you need to set it to True, otherwise default value is False, to speed up the parsing algorithm.
            messy_structure: If data is not organised properly in folders.
            image_names: List of names for NRRD files containing image file, default is ['image','volume','img','vol'].
        """

        self.index = 0
        self._data_path = data_path
        self._data_type = data_type
        self.__patients = os.listdir(data_path)
        self._patient_dict = OrderedDict() 
        if mask_names:
            self.__mask_names = mask_names
        else:
            self.__mask_names = ['mask']
        
        self._image_only = image_only
        self.__image_names = image_names
        self.__multi_rts_per_pat = multi_rts_per_pat
        self.__parse_directory(messy_structure)
        
        
    def __parse_directory(self, messy_structure=False):
        
        if self._data_type =='nrrd':
            
            for patient in tqdm(self.__patients):
                temp_files = []
                for root, dirs, files in os.walk(os.path.join(self._data_path,patient)):
                    for file in files:
                        if file.endswith('.nrrd'):
                            temp_files.append(os.path.join(root,file))
                if not len(temp_files):
                    raise FileNotFoundError('No nrrd/mha data found: check the folder')
            
                for mask_name in self.__mask_names:         
                    for file in temp_files:
                        if is_in_list(self.__image_names,file.split('\\')[-1][:-5]):
                            if not self._image_only:
                                for mfile in temp_files:
                                    if re.search(mask_name.lower(),mfile.lower()):
                                        self._patient_dict[str(patient+'_'+ mfile.split('\\')[-1][:-5])]=[file,mfile] #[-20:-5]
                            else:
                                self._patient_dict[str(patient)]=[file]    
            
        elif self._data_type =='dcm':
            for patient in tqdm(self.__patients):
            
                dcm_files=[]
                all_dicom_subfiles = []
                dict_length = len(self._patient_dict)

                if messy_structure:
                    for root, dirs, files in os.walk(os.path.join(self._data_path)):
                        for file in files:
                            if file.endswith('.dcm'):
                                temp_file = pydicom.read_file(os.path.join(root, file),force = True)
                                if temp_file.Modality == 'RTSTRUCT':
                                    all_dicom_subfiles.append(os.path.join(root, file))
                    print(len(all_dicom_subfiles))
                
                for root, dirs, files in os.walk(os.path.join(self._data_path,patient)):
                    for file in files:
                        if file.endswith('.dcm'):
                            dcm_files.append(os.path.join(root, file))
                            
                if not len(dcm_files):
                    warn('No dcm data found for patient:%s check the folder, ensure that dicom files ends with .dcm'%patient)

                if messy_structure:
                    for rt in all_dicom_subfiles:
                        temp_rt = pydicom.read_file(rt, force=True)
                        for dfile in dcm_files:
                            temp_dfile = pydicom.read_file(dfile, force=True)
                            if is_in_list(['ct', 'mr', 'us', 'nm'], str(temp_dfile.Modality).lower()):
                                if (temp_dfile.SeriesInstanceUID != temp_rt.SeriesInstanceUID):
                                    pass
                                else:
                                    datafile = os.path.dirname(os.path.abspath(dfile))
                                    rts_name = rt[:-4].split('\\')[-2]  # +rt[:-4].split('\\')[-2]
                                    self._patient_dict[patient + '_' + rts_name[-15:]] = [datafile, rt]
                                    break
                else:
                    for file in dcm_files:
                        structfile = ''
                        try:
                            temp_file = pydicom.read_file(file, force=True)

                            if not self._image_only:
                                if temp_file.Modality == 'RTSTRUCT':
                                    structfile = file
                                    datafile = ''
                                    for dfile in dcm_files:
                                        temp_dfile = pydicom.read_file(dfile, force=True)

                                        if is_in_list(['ct', 'mr', 'us', 'nm'], str(temp_dfile.Modality).lower()):
                                            if (temp_dfile.StudyInstanceUID != temp_file.StudyInstanceUID):
                                                warn('StudyInstanceUID doesnt match!')

                                            datafile = os.path.dirname(os.path.abspath(dfile))
                                            rts_name = structfile[:-4].split('\\')[-1]
                                            self._patient_dict[patient + '_' + rts_name[-15:]] = [datafile, structfile]
                                            break

                            else:
                                datafile = ''
                                for dfile in dcm_files:
                                    temp_dfile = pydicom.read_file(dfile, force=True)
                                    if is_in_list(['ct', 'mr', 'us', 'nm'], str(temp_dfile.Modality).lower()):
                                        datafile = os.path.dirname(os.path.abspath(dfile))  # getting directory name
                                        self._patient_dict[patient] = [datafile]
                                        break

                        except KeyboardInterrupt:
                            raise
                        except:
                            warn('Some problems have occured with the file: %s' % file)

                        if not self.__multi_rts_per_pat and dict_length - len(self._patient_dict):
                            break
                
      
        else:
            raise NotImplementedError('Currently only "dcm" format and "nrrd" (nrrd/mha) formats are supported')
      
    def __iter__(self):
        return self
    def __len__(self):
        return(len(self._patient_dict))
    
    def __next__(self):
        if self.index == len(self._patient_dict):
            self.index = 0
            raise StopIteration
        temp_key = [*self._patient_dict.keys()][self.index]
        temp_data = self._patient_dict[temp_key]
        self.index +=1
        return temp_key,temp_data
    




# -*- coding: utf-8 -*-
"""
Created on Tue Jun 6 13:52:52 2020
@author: E.Lavrova
e.lavrova@maastrichtuniversity.nl
"""

import pandas as pd
import numpy as np
import plotly.graph_objs as go
import plotly
import plotly.subplots
from statsmodels.stats import multitest
import os
import scipy as sp
import math
import scipy.stats
from operator import le, gt
from sklearn.metrics import auc
import plotly.io as pio


class FeaturesSet:
    '''This class creates features set objects.'''

    def __init__(self,
                 feature_path: str =None,
                 outcome_path: str =None,
                 feature_column: list =[],
                 feature_column_to_drop: list =[],
                 outcome_column: str ='',
                 patient_column: str ='',
                 patient_in_outcome_column: str ='',
                 patient_to_drop: list =[]):

        """Initialise a dataset object.

        Arguments:
            feature_path: Path to csv./.xls(x) file with features.
            outcome_path: Path to csv./.xls(x) file with outcomes.
            feature_column: List of features to be included.
            feature_column_to_drop: List of features to be excluded.
            outcome_column: Name of the column with outcomes.
            patient_column: Name of the column with patients IDs in features file.
            patient_in_outcome_column: Name of the column with patients IDs in outcomes file.
            patient_to_drop: List of patients to be excluded.
        """

        if type(feature_path) is str:
            self._feature_path = feature_path
        else:
            print('Features csv/xls file path has wrong format.')
        if type(patient_column) is str:
            self._patient_column = patient_column
        else:
            print('Patient column name has wrong format.')
        if type(outcome_path) is str:
            self._outcome_path = outcome_path
        else:
            print('Outcome csv/xls file path has wrong format.')
        if type(feature_column) is list:
            self._feature_column = feature_column
        else:
            print('List of feature columns has wrong format.')
        if type(outcome_column) is str:
            self._outcome_column = outcome_column
        else:
            print('Outcome column name has wrong format.')
        if type(patient_in_outcome_column) is str:
            self._patient_in_outcome_column = patient_in_outcome_column
        else:
            print('Patient column name in dataframe with outcomes has wrong format.')
        if type(patient_to_drop) is list:
            self._patient_to_drop = patient_to_drop
        else:
            print('List of patient names to be excluded has wrong format.')
        if type(feature_column_to_drop) is list:
            self._feature_column_to_drop = feature_column_to_drop
        else:
            print('List of feature names to be excluded has wrong format.')
        self._class_label = []
        self._outcome = []
        self._feature_dataframe = None
        self._feature_outcome_dataframe = None
        if len(feature_path)>0 and len(patient_column)>0:
            self.__read_files()
        else:
            print('Path to csv/xls with features or patient column name is missing.')

    def __read_files(self):

        # reads .csv/.xls(x) tables with features and outcomes and gets feature_set attributes

        if '.csv' in self._feature_path:
            feature_df = pd.read_csv(self._feature_path, dtype={self._patient_column: str})
        elif '.xls' in self._feature_path:
            feature_df = pd.read_excel(self._feature_path, dtype={self._patient_column: str})
        else:
            print('Data format is not supported')
            return

        if len(self._feature_column) > 0:
            self._feature_column = list(set(self._feature_column) & set(list(feature_df.columns)))
        else:
            self._feature_column = list(feature_df.columns)

        if len(self._outcome_column) > 0:
            if self._outcome_column in self._feature_column:
                self._feature_column.remove(self._outcome_column)

        if len(self._feature_column_to_drop) > 0:
            for feature in self._feature_column_to_drop:
                if feature in self._feature_column:
                    self._feature_column.remove(feature)

        if '' in self._feature_column:
            self._feature_column.remove('')
        if 'Unnamed: 0' in self._feature_column:
            self._feature_column.remove('Unnamed: 0')

        technical_features_to_remove = []
        for feature in self._feature_column:
            if ('diagnostics' in feature) or ('general' in feature):
                technical_features_to_remove.append(feature)
        for feature in technical_features_to_remove:
            self._feature_column.remove(feature)

        if len(self._patient_column) > 0:
            if self._patient_column in list(feature_df.columns):
                self._patient_name = list(feature_df[self._patient_column])
                feature_df.set_index(self._patient_column, inplace=True)
                if len(self._patient_to_drop) > 0:
                    for patient in self._patient_to_drop:
                        if patient in self._patient_name:
                            self._patient_name.remove(patient)
                feature_df = feature_df.reindex(self._patient_name)
            if self._patient_column in self._feature_column:
                self._feature_column.remove(self._patient_column)

        self._feature_dataframe = feature_df[self._feature_column].copy()

        if len(self._outcome_path) > 0:
            if (len(self._patient_column) > 0) & (len(self._patient_in_outcome_column) > 0):
                if '.csv' in self._outcome_path:
                    outcome_df = pd.read_csv(self._outcome_path, dtype={self._patient_in_outcome_column: str,
                                                                        self._outcome_column: str})
                elif '.xls' in self._outcome_path:
                    outcome_df = pd.read_excel(self._outcome_path, dtype={self._patient_in_outcome_column: str,
                                                                          self._outcome_column: str})
                outcome_df.set_index(self._patient_in_outcome_column, inplace=True)
                self._outcome = outcome_df[self._outcome_column]

        else:
            if self._outcome_column in list(feature_df.columns):
                self._outcome = feature_df[self._outcome_column]

        if len(self._outcome) > 0:
            self._feature_outcome_dataframe = self._feature_dataframe.copy()
            self._feature_outcome_dataframe[self._outcome_column] = None
            for patient_outcome in list(self._outcome.index):
                for patient in self._patient_name:
                    if patient_outcome in patient:
                        self._feature_outcome_dataframe.at[patient, self._outcome_column] = \
                            self._outcome[patient_outcome]
            self._outcome = self._feature_outcome_dataframe[self._outcome_column]
            self._class_label = pd.unique(np.array(list(self._feature_outcome_dataframe[self._outcome_column])))
            self._class_label.sort()
            data_balance = []
            for l in self._class_label:
                data_balance.append(np.sum(np.array(list(self._outcome)) == l)/len(self._outcome))

        print('Number of observations: {}\nClass labels: {}\nClasses balance: {}'.format(len(self._outcome),
                                                                                  self._class_label,
                                                                                  data_balance))
        return None















# -*- coding: utf-8 -*-
"""
Created on Thu Apr 23 20:13:55 2020
@author: S.Primakov
s.primakov@maastrichtuniversity.nl
"""

import os,re,logging
from pmtool.DataSet import DataSet
import pydicom
import numpy as np
from skimage import draw
import SimpleITK as sitk
from warnings import warn
from tqdm import tqdm
import matplotlib.pyplot as plt
import radiomics
from radiomics import featureextractor
from pandas import DataFrame
import pandas as pd
import csv
import cv2

class MyException(Exception):
    pass

class ToolBox(DataSet):

    '''This module is inherited from DataSet class and allows for high-level functionality while working with the raw imaging data.'''

    def get_dataset_description(self, parameter_list: list =['Modality', 'SliceThickness',
                    'PixelSpacing', 'SeriesDate', 'Manufacturer']) -> DataFrame:
        """Get specified metadata from the DICOM files of the dataset.

        Arguments:
            parameter_list: List of the parameters to be collected from DICOM files.

        Returns:
            DataFrame with the specified parameters for each patient.
        """
        CT_params = ['PatientName','ConvolutionKernel','SliceThickness',
                      'PixelSpacing','KVP','Exposure','XRayTubeCurrent',
                      'SeriesDate']
        MRI_params = ['Manufacturer','SliceThickness','PixelSpacing',
                      'StudyDate','MagneticFieldStrength','EchoTime']

        if self._data_type =='dcm':
            if parameter_list == 'MRI':
                params_list = MRI_params

            elif parameter_list == 'CT':
                params_list = CT_params
            else:
                params_list = parameter_list

            dataset_stats = DataFrame(data=None, columns = params_list )
            for pat,path in tqdm(self, desc='Patients processed'):
                image,_ = self.__read_scan(path[0])
                for i,temp_slice in enumerate(image):
                    dataset_stats = dataset_stats.append(pd.Series([pat,str(i),*[self.__val_check(temp_slice,x) for x in params_list]],index = ['patient','slice#',*params_list]),ignore_index=True)

            return dataset_stats
        else:
            warn('Only available for DICOM dataset')

    def get_jpegs(self, export_path: str):
        '''Convert each slice of nnrd containing ROI to JPEG image. Quick and convienient way to check your ROIs after conversion.

        Arguments:
            export_path: Path to the folder where the JPEGs will be generated.
        '''
        if self._data_type =='nrrd':
            for pat,path in tqdm(self, desc='Patients processed'):
                try:
                    temp_data = sitk.ReadImage(path[0])
                    temp_mask = sitk.ReadImage(path[1])
                    temp_image_array = sitk.GetArrayFromImage(temp_data)
                    temp_mask_array = sitk.GetArrayFromImage(temp_mask)

                    directory = os.path.join(export_path,'images_quick_check',pat,path[1][:-5].split('\\')[-1])
                    z_dist = np.sum(temp_mask_array,axis = (1,2))
                    z_ind = np.where(z_dist!=0)[0]

                    for j in z_ind:
                        if not os.path.exists(directory):
                            os.makedirs(directory)
                        temp_image_array[j,0,0]= 1
                        temp_mask_array [j,0,0]= 1
                        plt.figure(figsize=(20,20))
                        plt.subplot(121)
                        plt.imshow(temp_image_array[j,...],cmap = 'bone')
                        plt.subplot(122)
                        plt.imshow(temp_image_array[j,...],cmap = 'bone')
                        plt.contour(temp_mask_array[j,...],colors = 'red',linewidths = 2)#,alpha=0.7)
                        plt.savefig(os.path.join(directory,'slice #%d'%j),bbox_inches='tight')
                        plt.close()
                except:
                    warn('Something wrong with %s'%pat)

        else:
            raise TypeError('The toolbox should be initialized with data_type = "nrrd" ')

    def extract_features(self, params_file: str, loggenabled: bool =False) -> DataFrame:
        """Extract PyRadiomic features from the dataset.

                Arguments:
                    params_file: File with the PyRadiomics parameters for features extraction.
                    loggenabled: Enable/disable log file writing.

                Returns:
                    DataFrame with the extracted features.
        """

        if self._data_type =='nrrd':
            #set up pyradiomics
            if loggenabled:
                logger = radiomics.logger
                logger.setLevel(logging.DEBUG)  # set level to DEBUG to include debug log messages in log file

                # Write out all log entries to a file
                handler = logging.FileHandler(filename='test_log.txt', mode='w')
                formatter = logging.Formatter("%(levelname)s:%(name)s: %(message)s")
                handler.setFormatter(formatter)
                logger.addHandler(handler)

            # Initialize feature extractor using the settings file
            feat_dictionary,key_number={},0
            extractor = featureextractor.RadiomicsFeatureExtractor(params_file)

            for pat,path in tqdm(self,desc='Patients processed'):
                try:
                    temp_data = sitk.ReadImage(path[0])
                    temp_mask = sitk.ReadImage(path[1])
                    pat_features = extractor.execute(temp_data, temp_mask)

                    if pat_features['diagnostics_Image-original_Hash'] != '':
                        pat_features.update({'Patient':pat,'ROI':path[1].split('\\')[-1][:-5]})
                        feat_dictionary[key_number] = pat_features
                        key_number+=1

                except KeyboardInterrupt:
                    raise
                except:
                    warn('region : %s skipped'%pat)

            output_features = DataFrame.from_dict(feat_dictionary).T

            return output_features
        else:
            raise TypeError('The toolbox should be initialized with data_type = "nrrd"')

    def convert_to_nrrd(self, export_path: str, region_of_interest: str = 'all', image_type=np.int16):
        '''Convert DICOM dataset to the volume (NRRD) format.

        Arguments:
            export_path: Path to the folder where the converted NRRDs will be placed.
            region_of_interest: If you know exact name of the ROI you want to extract, then write it with the ! character in front, eg. region_of_interest = !gtv1 , if you want to extract all the GTVs in the rtstructure eg. gtv1, gtv2, gtv_whatever, then just specify the stem word eg. region_of_interest = gtv, default value is region_of_interest ='all' , which means that all ROIs in rtstruct will be extracted.
            image_type: Data type of the input image.
        '''
        if self._data_type == 'dcm':

            if self._image_only:

                for pat, pat_path in tqdm(self, desc='Patients converted'):
                    img_path = pat_path[0]
                    image = self.__get_image(img_path, image_type)

                    export_dir = os.path.join(export_path, 'converted_nrrds', pat)
                    if not os.path.exists(export_dir):
                        os.makedirs(export_dir)

                    image_file_name = 'image.nrrd'
                    sitk.WriteImage(image, os.path.join(export_dir, image_file_name))

            else:

                for pat, pat_path in tqdm(self, desc='Patients converted'):
                    img_path = pat_path[0]
                    rt_path = pat_path[1]
                    try:
                        rt_structure, roi_list = self.__get_roi(region_of_interest, rt_path)
                    except KeyboardInterrupt:
                        raise
                    except:
                        roi_list = []
                        warn('Error: ROI extraction failed for patient%s' % pat)

                    for roi in roi_list:
                        try:
                            image, mask = self.__get_binary_mask(img_path, rt_structure, roi, image_type)

                            export_dir = os.path.join(export_path, 'converted_nrrds', pat)

                            if not os.path.exists(export_dir):
                                os.makedirs(export_dir)

                            image_file_name = 'image.nrrd'
                            mask_file_name = '%s_mask.nrrd' % roi
                            sitk.WriteImage(image, os.path.join(export_dir,
                                                                image_file_name))  # save image and binary mask locally
                            sitk.WriteImage(mask, os.path.join(export_dir, mask_file_name))
                        except KeyboardInterrupt:
                            raise
                        except:
                            warn('Patients %s ROI : %s skipped' % (pat, roi))


        else:
            raise TypeError('Currently only conversion from dicom -> nrrd is available')

    def pre_process(self, ref_img_path: str = None, save_path: str = None,
                    z_score: bool = False, norm_coeff: tuple = None, hist_match: bool = False,
                    hist_equalize: bool = False, binning: bool = False, percentile_scaling: bool = False,
                    corr_bias_field: bool = False,
                    subcateneus_fat: bool = False, fat_value: bool = None,
                    reshape: bool = False, to_shape: np.array = None,
                    window_filtering_params: tuple = None,
                    verbosity: bool = False, visualize: bool = False):
        '''Pre-process the images.

        Arguments:
            ref_img_path: Path to the reference image for the histogram matching.
            save_path: Path to the folder where the pre-processed images will be stored.
            z_score: Enable Z-scoring.
            norm_coeff: Normalization coefficients for z-scoring (mean and standard deviation).
            hist_match: Enable histogram matching.
            hist_equalize: Enable histogram equalization.
            binning: Enable intensities resampling.
            percentile_scaling: Enable intensities scaling based on 95 percentile.
            corr_bias_field: Enable bias field correction.
            subcateneus_fat: Enable intensities rescaling based on subcateneus fat.
            fat_value: Intensity value specific for the fat.
            reshape: Enable 3D reshaping of the images.
            to_shape: Target shape for image reshaping.
            verbosity: Enable log reporting.
            visualize: Enable visualization of every pre-processing step.
        '''

        ref_img_arr = sitk.GetArrayFromImage(sitk.ReadImage(ref_img_path))
        for i, pat in tqdm(self):
            image = sitk.ReadImage(pat[0])
            mask = sitk.ReadImage(pat[1])
            image_array = sitk.GetArrayFromImage(image)
            mask_array = sitk.GetArrayFromImage(mask)
            ref_image_arr = ref_img_arr.copy()
            # run the pre-processing
            pre_processed_arr = self.__preprocessing_function(image_array, mask_array, ref_image_arr,
                                                              z_score, norm_coeff, hist_match, hist_equalize,
                                                              binning, percentile_scaling,
                                                              corr_bias_field, window_filtering_params,
                                                              subcateneus_fat,
                                                              fat_value,
                                                              reshape, to_shape,
                                                              verbosity, visualize)

            export_dir = os.path.join(save_path, i)
            if not os.path.exists(export_dir):
                os.makedirs(export_dir)

            pre_processed_image = sitk.GetImageFromArray(pre_processed_arr)
            pre_processed_image.CopyInformation(image)
            sitk.WriteImage(pre_processed_image,
                            os.path.join(export_dir, 'pre_processed_img.nrrd'))  # save image and binary mask locally
            sitk.WriteImage(mask, os.path.join(export_dir, 'mask.nrrd'))

        params = {'ref_img_path': ref_img_path, 'save_path': save_path, 'z_score': z_score, 'norm_coeff': norm_coeff,
                  'hist_match': hist_match, 'hist_equalize': hist_equalize, 'binning': binning,
                  'percentile_scaling': percentile_scaling, 'subcateneus_fat': subcateneus_fat, 'fat_value': fat_value,
                  'verbosity': verbosity, 'visualize': visualize}

        with open(os.path.join(save_path, 'pre-processing_parameters.csv'), 'w') as csv_file:
            writer = csv.writer(csv_file)
            writer.writerow(['Pre-processing parameters:'])
            for key, value in params.items():
                writer.writerow([str(str(key) + ': ' + str(value))])

    def get_quality_checks(self, qc_parameters: dict = {'specific_modality': '',
                                                        'thickness_range': [],
                                                        'scan_length_range': [],
                                                        'axial_res': [],
                                                        'spacing_range': [],
                                                        'kernels_list': []},
                           verbosity: bool =False) -> DataFrame:

        '''Perform a basic quality check for the data. If one of the quality checked parameters is not passed, the correspondig check is not to be performed.

                Arguments:
                    specific_modality str: Target modality of the dataset.
                    thickness_range list: Range of the acceptable slice thicknesses, minimum and maximum.
                    scan_length_range list: Range of the acceptable number of slices, minimum and maximum.
                    axial_res list: Target axial resolution, in both dimensions.
                    spacing_range list: Range of the acceptable in-plane pixel spacing, minimum and maximum.
                    kernels_list list: List of the acceptable kernels for CT data.
                    verbosity: Enable log reporting.
                Returns:
                    DataFrame with the quality check results for each patient. If '1' - check passed, if '0' - check failed.
                '''
        ## init dataframe, parse qa tags
        df_columns = ['Patient id', 'Modality is acceptable', 'Projection is axial',
                      'Complete scan (no missing/overlapping slices)',
                      'Scan len is in range', 'Slice thickness is in range',
                      'Slice thickness is consistent', 'Pixel spacing is in range',
                      'Convolutional kernel tag is present', 'Convolutional kernel is acceptable',
                      'Axial pr. resolution is acceptable', 'Intensity intercept/slope tags are present']

        checks_df = pd.DataFrame([], columns=df_columns)
        for pat in tqdm(self):
            checks = self.__quality_checks(*pat, qc_parameters=qc_parameters, columns=df_columns, verbosity=verbosity)
            checks_df = checks_df.append(checks, ignore_index=True)
        return checks_df

    def __quality_checks(self, patient, path, qc_parameters, columns, verbosity):
        scans, skipped_files = self.__read_scan(path[0])
        scans_sorted = False
        try:
            scans.sort(key=lambda x: x.ImagePositionPatient[2])
            scans_sorted = True
        except:
            warn('Some problems with sorting scans for pat:%s' % patient)
        img = scans[0]

        ##check scans modality
        if qc_parameters.get('specific_modality', 0):
            if len(qc_parameters['specific_modality']) > 0:
                try:
                    modality = self.__check_modality(scans, qc_parameters['specific_modality'])
                    if verbosity:
                        print('Modality check status:', modality)
                except:
                    modality = np.nan
                    print('Cannot perform Modality check for pat: %s' % patient)
            else:
                raise MyException(
                    "Please correctly specify the 'specific_modality' in parameters (e.g. 'specific_modality':'CT')")
        else:
            modality = '-'
            if verbosity:
                print("Acceptable value was not provided for the 'specific_modality', modality check was not performed ")

        ##number of slices_check
        if qc_parameters.get('scan_length_range', 0):
            if len(qc_parameters['scan_length_range']) == 2:
                try:
                    if len(scans) >= qc_parameters['scan_length_range'][0] \
                            and len(scans) <= qc_parameters['scan_length_range'][1]:
                        slice_nr = 1
                    else:
                        slice_nr = 0
                    if verbosity:
                        print('Number of slices check status:', slice_nr)
                except:
                    slice_nr = np.nan
                    print('Cannot perform slice # check for pat: %s' % patient)
            else:
                raise MyException("Please correctly specify the scan_length_range in parameters (e.g. 'scan_length_range':[1,50])")
        else:
            slice_nr = '-'
            if verbosity:
                print('Acceptable range was not provided for the number of slices check.')

        ##pixel_spacing range check
        if qc_parameters.get('spacing_range', 0):
            if len(qc_parameters['spacing_range']) == 2:
                try:
                    ps = img.PixelSpacing
                    if ps[0] == ps[1] and ps[0] >= qc_parameters['spacing_range'][0] and ps[0] <= qc_parameters['spacing_range'][1]:
                        pixel_sp = 1
                    else:
                        pixel_sp = 0
                    if verbosity:
                        print('Pixel spacing in acceptable range check status:', pixel_sp)
                except:
                    pixel_sp = np.nan
                    print('Cannot perform pixel spacing range check for pat: %s' % patient)
            else:
                raise MyException("Please correctly specify the spacing_range in parameters (e.g. 'spacing_range':[0.9,1.5])")
        else:
            pixel_sp = '-'
            if verbosity:
                print('Acceptable range was not provided for the pixel spacing range check.')

        ##check axial
        try:
            axial = self.__check_image_axial_plane(img)
            if verbosity:
                print('Axial plane check status:', axial)
        except:
            axial = np.nan
            print('Cannot perform axial plane check for pat: %s' % patient)

        ##check rescaling constants
        if qc_parameters.get('axial_res', 0):
            if len(qc_parameters['axial_res']) == 2:
                try:
                    intercept_slope, image_shape = self.__get_pixel_values(scans,np.int8,True,qc_parameters['axial_res'])
                    if verbosity:
                        print('Intercept/slope check status:', intercept_slope)
                        print('Image shape is %s check status'%qc_parameters['axial_res'], image_shape)
                except:
                    intercept_slope, image_shape = np.nan, np.nan
                    print('Cannot perform Intercept/slope and image shape check for pat: %s' % patient)
            else:
                raise MyException("Please correctly specify the axial_res in parameters (e.g. 'axial_res':[512,512])")
        else:
            intercept_slope, image_shape = '-', '-'
            if verbosity:
                print('Acceptable range was not provided for the axial resolution, axial pr. resolution check & Intercept/slope check are not performed ')

        ##check slice thickness consistency
        if qc_parameters.get('thickness_range', 0):
            if len(qc_parameters['thickness_range']) == 2:
                try:
                    slice_thickness_consistency, slice_thickness_range = self.__check_consistent_slice_thickness(scans,
                                                                                                              thickness_range=qc_parameters['thickness_range'])
                    if verbosity:
                        print('Slice thickness consistency check status:', slice_thickness_consistency)
                        print('Slice thickness in acceptable range check status:', slice_thickness_range)
                except:
                    slice_thickness_consistency, slice_thickness_range = np.nan, np.nan
                    print('Cannot perform slice thickness consistency and range check for pat: %s' % patient)
            else:
                raise MyException("Please correctly specify the 'thickness_range' in parameters (e.g. 'thickness_range':[3,5])")
        else:
            slice_thickness_consistency, slice_thickness_range = '-', '-'
            if verbosity:
                print('Acceptable range was not provided for the thickness_range, slice thickness range check & slice thickness consistency check are not performed ')

        ##check missing overlaping slices
        try:
            missing_overlapping_slices = self.__check_missing_overlapping_slices(scans)
            if verbosity:
                print('Missing/overlapping slices check status:', missing_overlapping_slices)
        except:
            missing_overlapping_slices = np.nan
            print('Cannot perform missing/overlapping slice check for pat: %s' % patient)

        ##check conv kernel
        if qc_parameters.get('kernels_list', 0):
            if len(qc_parameters['kernels_list']) >0:
                try:
                    conv_kern = self.__check_kernel(scans, qc_parameters['kernels_list'])
                    conv_pres = 1
                    if verbosity:
                        print('Conv Kernel check status:', conv_kern)
                except:
                    conv_kern = np.nan
                    conv_pres = 0
                    print('Cannot perform conv kernel check for pat: %s' % patient)
            else:
                raise MyException("Please correctly specify the 'kernels_list' in parameters (e.g. 'kernels_list':['standard','lung'])")
        else:
            conv_kern, conv_pres = '-', '-'
            if verbosity:
                print("Acceptable value was not provided for the 'kernels_list', conv kernel check was not performed ")

        checks = [patient, modality, axial, missing_overlapping_slices, slice_nr, slice_thickness_range,
                  slice_thickness_consistency, pixel_sp, conv_pres, conv_kern,
                  image_shape, intercept_slope]
        return pd.Series(checks, index=columns)

    def __check_image_axial_plane(self, img):
        if int(img.ImageOrientationPatient[0]) and int(img.ImageOrientationPatient[4]) and not int(
                img.ImageOrientationPatient[1]) and not int(img.ImageOrientationPatient[2]) and not int(
                img.ImageOrientationPatient[3]) and not int(img.ImageOrientationPatient[5]):
            return 1
        else:
            return 0

    def __check_modality(self,scans,speciefic_mod):
        mod_list = [str(x.Modality).lower() for x in scans]
        if mod_list.count(speciefic_mod.lower()) == len(mod_list):
            return 1
        else:
            return 0

    def __check_consistent_slice_thickness(self, scans, thickness_range):
        slice_thickness = [np.round(x.SliceThickness, 1) for x in scans]
        if slice_thickness.count(slice_thickness[0]) == len(slice_thickness):
            if slice_thickness[0] >= thickness_range[0] and slice_thickness[1] <= thickness_range[1]:
                return 1, 1
            else:
                return 1, 0
        else:
            if slice_thickness[0] >= thickness_range[0] and slice_thickness[1] <= thickness_range[1]:
                return 0, 1
            else:
                return 0, 0

    def __check_missing_overlapping_slices(self, scans):
        temp_spacing = []
        scan_real_range = np.round(list(np.float(x.ImagePositionPatient[2]) for x in scans), 1)
        for i in range(len(scan_real_range) - 1):
            temp_spacing.append(np.round(scan_real_range[i + 1] - scan_real_range[i], 1))

        if temp_spacing.count(temp_spacing[0]) == len(temp_spacing):
            return 1
        else:
            return 0

    def __check_kernel(self, scans, kernels=[]):
        kernels = [x.lower() for x in kernels]
        kernels_list = [x.ConvolutionKernel for x in scans]
        if kernels_list.count(kernels_list[0]) == len(kernels_list) and kernels_list[0].lower() in set(kernels):
            return 1
        else:
            return 0

    def __get_roi_id(self,rtstruct,roi):

        for i in range(len(rtstruct.StructureSetROISequence)):
            if str(roi)==rtstruct.StructureSetROISequence[i].ROIName:
                roi_number = rtstruct.StructureSetROISequence[i].ROINumber
                break
        for j in range(len(rtstruct.StructureSetROISequence)):
            if (roi_number==rtstruct.ROIContourSequence[j].ReferencedROINumber):
                break
        return j

    def __get_roi(self,region_of_interest,rt_path):

        rt_structure = pydicom.read_file(rt_path,force=True)
        roi_list = []
        if region_of_interest.lower()=='all':
            roi_list = [rt_structure.StructureSetROISequence[x].ROIName for x in range(0,len(rt_structure.StructureSetROISequence))]
        else:
            for i in range(0,len(rt_structure.StructureSetROISequence)):
                if region_of_interest.lower()[0]=='!':
                    exact_roi = region_of_interest[1:]
                    if exact_roi.lower()== str(rt_structure.StructureSetROISequence[i].ROIName).lower():
                        roi_list.append(rt_structure.StructureSetROISequence[i].ROIName)       ## only roi with the exact same name

                elif re.search(region_of_interest.lower(),str(rt_structure.StructureSetROISequence[i].ROIName).lower()):
                    roi_list.append(rt_structure.StructureSetROISequence[i].ROIName)

        return rt_structure,roi_list

    def __coordinates_to_mask(self,row_coords, col_coords, shape):
        mask = np.zeros(shape)
        pol_row_coords, pol_col_coords = draw.polygon(row_coords, col_coords, shape)
        mask[pol_row_coords, pol_col_coords] = 1
        return mask

    def __read_scan(self, path):
        scan = []
        skiped_files = []
        for s in os.listdir(path):
            try:
                temp_file = pydicom.read_file(os.path.join(path, s), force=True)
                temp_file.file_meta.TransferSyntaxUID = pydicom.uid.ImplicitVRLittleEndian
                temp_mod = temp_file.Modality
                scan.append(temp_file)
                if (temp_mod == 'RTSTRUCT') or (temp_mod == 'RTPLAN') or (temp_mod == 'RTDOSE'):
                    scan.remove(temp_file)
            except:
                skiped_files.append(s)

        try:
            scan.sort(key = lambda x: x.ImagePositionPatient[2])
        except:
            warn('Some problems with sorting scans')

        return scan,skiped_files

    def __get_pixel_values(self,scans,image_type,qa=False,image_res=[512,512]):
        try:
            image = np.stack([s.pixel_array*s.RescaleSlope+s.RescaleIntercept for s in scans])
            if qa:
                if image.shape[1] == image_res[0] and image.shape[2] == image_res[1]:
                    return 1, 1
                else:
                    return 1, 0
            else:
                return image.astype(image_type)
        except:
            warn('Problems occured with rescaling intensities')
            image = np.stack([s.pixel_array for s in scans])
            if qa:
                return 0, 0
            else:
                return image.astype(image_type)

    def __get_image(self, img_path,image_type):

        image, _ = self.__read_scan(img_path)
        img_first_slice = image[0]
        img_array = self.__get_pixel_values(image, image_type)

        xres = np.array(img_first_slice.PixelSpacing[0])
        yres = np.array(img_first_slice.PixelSpacing[1])
        zres = np.abs(image[1].ImagePositionPatient[2] - image[0].ImagePositionPatient[2])
        image_sitk = sitk.GetImageFromArray(img_array.astype(image_type))
        image_sitk.SetSpacing((float(xres), float(yres), float(zres)))
        image_sitk.SetOrigin(
            (float(img_first_slice.ImagePositionPatient[0]), float(img_first_slice.ImagePositionPatient[1]),
             float(img_first_slice.ImagePositionPatient[2])))


        return image_sitk

    def __get_binary_mask(self,img_path,rt_structure,roi,image_type):

        image,_ = self.__read_scan(img_path)

        precision_level = 0.5
        img_first_slice = image[0]
        img_array = self.__get_pixel_values(image,image_type)

        img_length=len(image)

        mask=np.zeros([img_length, img_first_slice.Rows, img_first_slice.Columns],dtype=np.uint8)
        xres=np.array(img_first_slice.PixelSpacing[0])
        yres=np.array(img_first_slice.PixelSpacing[1])
        zres=np.abs(image[1].ImagePositionPatient[2] - image[0].ImagePositionPatient[2])
        roi_id = self.__get_roi_id(rt_structure,roi)

        Fm = np.zeros((3,2))
        Fm[:,0],Fm[:,1] = img_first_slice.ImageOrientationPatient[:3],img_first_slice.ImageOrientationPatient[3:]
        row_pr,column_pr = img_first_slice.PixelSpacing[0],img_first_slice.PixelSpacing[1]

        T_vect = img_first_slice.ImagePositionPatient
        T_1 = np.array(img_first_slice.ImagePositionPatient)
        T_n = np.array(image[-1].ImagePositionPatient)
        k_column = (T_1 - T_n)/(1 - len(image))

        A_multi = np.zeros((4,4))
        A_multi[:3,0],A_multi[:3,1] = Fm[:,0]*row_pr,Fm[:,1]*column_pr
        A_multi[:3,3],A_multi[:3,2] = T_vect,k_column
        A_multi[3,3] = 1

        transform_matrix = np.linalg.inv(A_multi)

        for sequence in rt_structure.ROIContourSequence[roi_id].ContourSequence:
            temp_contour = sequence.ContourData
            contour_first_point = np.array([*sequence.ContourData[:3], 1])
            slice_position = np.dot(transform_matrix,contour_first_point)[2]

            if np.abs(np.round(slice_position)-slice_position)< precision_level:
                assert slice_position < len(image) and slice_position >= 0
                z_index = int(np.round(slice_position))
            else:
                warn('Cant find the slice position, try to select lower precison level')
                warn('Slice position is: ',slice_position,' suggested position is: ',np.round(slice_position))
                z_index = None

            x,y,z=[],[],[]

            for i in range(0,len(temp_contour),3):
                x.append(temp_contour[i+0])
                y.append(temp_contour[i+1])
                z.append(temp_contour[i+2])

            x,y,z=np.array(x),np.array(y),np.array(z)

            coord = np.vstack((x,y,z,np.ones_like(x)))

            pixel_coords_c_r_s = np.dot(transform_matrix,coord)
            img_slice = self.__coordinates_to_mask(pixel_coords_c_r_s[1,:],pixel_coords_c_r_s[0,:],[img_array.shape[1],img_array.shape[2]])

            assert z_index

            mask[z_index,:,:] = np.logical_or(mask[z_index,:,:],img_slice)

        image_sitk = sitk.GetImageFromArray(img_array.astype(image_type))
        mask_sitk = sitk.GetImageFromArray(mask.astype(np.int8))

        image_sitk.SetSpacing((float(xres),float(yres),float(zres)))
        mask_sitk.SetSpacing((float(xres),float(yres),float(zres)))
        image_sitk.SetOrigin((float(img_first_slice.ImagePositionPatient[0]),float(img_first_slice.ImagePositionPatient[1]),
                       float(img_first_slice.ImagePositionPatient[2])))
        mask_sitk.SetOrigin((float(img_first_slice.ImagePositionPatient[0]),float(img_first_slice.ImagePositionPatient[1]),
                       float(img_first_slice.ImagePositionPatient[2])))

        return image_sitk, mask_sitk

    def __val_check(self,file,value):
        try:
            val = getattr(file,value)
            if val == '' or val==' ':
                return 'NaN'
            else:
                return val
        except:
            return 'NaN'

    def __normalize_image_zscore(self, image, norm_coeff):  ##Zscore whole image
        img = image - norm_coeff[0]
        img = img / norm_coeff[1]
        return img

    def __normalize_image_zscore_per_image(self, image, mask, verbosity):  ##Zscore based on the masked region intensities
        image_s = image.copy()
        mu = np.mean(image_s.flatten())
        sigma = np.std(image_s.flatten())
        if verbosity:
            print('ROI MU = %s, sigma = %s' % (mu, sigma))
        img = (image - mu) / sigma
        return img, mu, sigma

    def __resample_intensities(self, orig_img, bin_nr, verbosity):  ## Intensity resampling whole image/or masked region
        v_count = 0
        img_list = []
        filtered = orig_img.copy()
        if np.min(orig_img.flatten()) < 0:
            filtered += np.abs(np.min(orig_img.flatten()))
        resampled = np.zeros_like(filtered)
        max_val_img = np.max(filtered.flatten())
        step = max_val_img / (1.0*bin_nr)
        print('min: ',np.min(filtered.flatten()),'max: ',np.max(filtered.flatten()),'step: ',step)
        for st in np.arange(step, max_val_img + step, step):
            resampled[(filtered <= st) & (filtered >= st - step)] = v_count
            v_count += 1
        if verbosity:
            print("Resampling with a step of: ", step, 'Amount of unique values, original img: ',
                  len(np.unique(orig_img.flatten())), 'resampled img: ', len(np.unique(resampled.flatten())))

        if len(np.unique(resampled.flatten())) < 255:
            return np.array(resampled, dtype=np.uint8)
        else:
            return np.array(resampled, dtype=np.uint16)

    def __intensity_scaling(self, img, fat_int_value=None, method=''):
        if method == 'fat' and fat_int_value:
            filtered = img.copy()
            if np.min(img.flatten()) < 0:
                filtered += np.min(img.flatten())
            filtered = (1.0 * filtered) / (1.0 * np.max(img.flatten()))
            filtered = 1.0 * fat_int_value * filtered
            return np.array(filtered, np.uint16)

        elif method == '95th':
            filtered = img.copy()
            if np.min(img.flatten()) < 0:
                filtered += np.min(img.flatten())
            perc_95 = np.percentile(filtered.flatten(), 95)
            filtered = (1.0 * filtered) / (1.0 * np.max(img.flatten()))
            filtered = 1.0 * perc_95 * filtered
            return np.array(filtered, np.uint16)
        else:
            print('max value is not understood, skipping intensity rescaling step!')

    def __histogram_matching(self, source, template):
        """
        Adjust the pixel values of a grayscale image such that its histogram
        matches that of a target image

        Arguments:
        -----------
            source: np.ndarray
                Image to transform; the histogram is computed over the flattened
                array
            template: np.ndarray
                Template image; can have different dimensions to source
        Returns:
        -----------
            matched: np.ndarray
                The transformed output image
        """

        oldshape = source.shape
        source = source.ravel()
        template = template.ravel()

        # get the set of unique pixel values and their corresponding indices and
        # counts
        s_values, bin_idx, s_counts = np.unique(source, return_inverse=True,
                                                return_counts=True)
        t_values, t_counts = np.unique(template, return_counts=True)

        # take the cumsum of the counts and normalize by the number of pixels to
        # get the empirical cumulative distribution functions for the source and
        # template images (maps pixel value --> quantile)
        s_quantiles = np.cumsum(s_counts).astype(np.float64)
        s_quantiles /= s_quantiles[-1]
        t_quantiles = np.cumsum(t_counts).astype(np.float64)
        t_quantiles /= t_quantiles[-1]

        # interpolate linearly to find the pixel values in the template image
        # that correspond most closely to the quantiles in the source image
        interp_t_values = np.interp(s_quantiles, t_quantiles, t_values)

        return interp_t_values[bin_idx].reshape(oldshape).astype(np.int16)

    def __resize_3d_img(self, img, shape, interp=cv2.INTER_CUBIC):

        init_img = img.copy()
        temp_img = np.zeros((init_img.shape[0], shape[1], shape[2]))
        new_img = np.zeros(shape)
        for i in range(0, init_img.shape[0]):
            temp_img[i, ...] = cv2.resize(init_img[i, ...], dsize=(shape[2], shape[1]), interpolation=interp)
        for j in range(0, shape[1]):
            new_img[:, j, :] = (cv2.resize(temp_img[:, j, :], dsize=(shape[2], shape[0]), interpolation=interp))
        return new_img

    def __correct_bias_field(self, img, mask, n_fitting_levels, n_iterations):

        img_sitk = sitk.GetImageFromArray(img)
        mask_sitk = sitk.GetImageFromArray(mask)

        img_sitk_cast = sitk.Cast(img_sitk, sitk.sitkFloat32)
        mask_sitk_cast = sitk.Cast(mask_sitk, sitk.sitkUInt8)

        corrector = sitk.N4BiasFieldCorrectionImageFilter()
        corrector.SetMaximumNumberOfIterations(n_fitting_levels * n_iterations)

        img_output = sitk.GetArrayFromImage(corrector.Execute(img_sitk_cast, mask_sitk_cast))

        return img_output

    def __apply_window(self, img, window_params, verbosity=False):
        new_img = img.copy()
        WW = window_params[0]
        WL = window_params[1]
        assert (WW < 2000) and (WW > 0) and (WL < 200) and (WL > -1000)
        up_lim, low_lim = WL + WW / 2, WL - WW / 2
        if low_lim < -1000:
            low_lim = -1000
        if verbosity:
            print('Window limits ', low_lim, up_lim)
        new_img[np.where(img < low_lim)] = low_lim
        new_img[np.where(img > up_lim)] = up_lim
        return new_img

    def __preprocessing_function(self, img, mask, ref_img, z_score, norm_coeff, hist_match, hist_equalize, binning,
                                 percentile_scaling, corr_bias_field, window_filtering_params,
                                 subcateneus_fat, fat_value, reshape, to_shape,
                                 verbosity, visualize):

        if verbosity:
            unique_number_of_intensity = np.unique(img.flatten())
            min_int, max_int = np.min(img.flatten()), np.max(img.flatten())
            img_type = img.dtype
            print('-' * 40)
            print('Original image stats:')
            print('Number of unique intensity values :', len(unique_number_of_intensity))
            print('Intensity values range:[%s:%s]' % (min_int, max_int))
            print('Image intensity dtype:', img_type)
            print('-' * 40)
        if visualize:
            plt.figure(figsize=(12, 12))
            plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
            plt.title('original image')
            plt.show()

        if corr_bias_field:
            img = self.__correct_bias_field(img, mask, n_fitting_levels=4, n_iterations=[50])
            if verbosity:
                print('N4 bias field correction performed')
            if visualize:
                plt.figure(figsize=(12, 12))
                plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
                plt.title('Bias field correction')
                plt.show()

        if len(window_filtering_params) == 2:
            img = self.__apply_window(img, window_filtering_params,verbosity)
            if verbosity:
                print('Window filtering applied')
            if visualize:
                plt.figure(figsize=(12, 12))
                plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
                plt.title('Window filtering')
                plt.show()

        if subcateneus_fat:
            img = self.__intensity_scaling(np.squeeze(img), fat_value, method='fat')
            if verbosity:
                print('Intensity values rescaled based on the subcateneus fat')
            if visualize:
                plt.figure(figsize=(12, 12))
                plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
                plt.title('subcateneus fat scaling')
                plt.show()

        if percentile_scaling:
            img = self.__intensity_scaling(np.squeeze(img), method='95th')
            if verbosity:
                print('Intensity values rescaled based on the 95th percentile')
            if visualize:
                plt.figure(figsize=(12, 12))
                plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
                plt.title('95th percentile scaling')
                plt.show()

        if hist_match:
            img = self.__histogram_matching(np.squeeze(img), np.squeeze(ref_img))
            if verbosity:
                print('Histogram matching was applied')
            if visualize:
                plt.figure(figsize=(22, 12))
                plt.subplot(121)
                plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
                plt.title('Histogram matched')
                plt.subplot(122)
                plt.imshow(ref_img[int(len(img) / 2.), ...], cmap='bone')
                plt.title('Ref image')
                plt.show()

        if binning:
            img = self.__resample_intensities(img, binning, verbosity)
            if verbosity:
                print('-' * 40)
                print('Intensity values resampled with a number of bins: %s' % binning)
                print('-' * 40)
            if visualize:
                plt.figure(figsize=(12, 12))
                plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
                plt.title('%s bins resampled' % binning)
                plt.show()

        if hist_equalize and img.dtype == np.uint8:
            equalized = [cv2.equalizeHist(np.squeeze(x).astype(np.uint8)) for x in img]
            img = np.stack(equalized)
            if verbosity:
                print('Histogram equalization applied')
            if visualize:
                plt.figure(figsize=(12, 12))
                plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
                plt.title('Histogram equalize')
                plt.show()

        if z_score and norm_coeff:
            img = self.__normalize_image_zscore(img, norm_coeff)
            if verbosity:
                print('Z-score normalization applied based on the whole image, Mu=%s, sigma=%s' % (norm_coeff))
            if visualize:
                plt.figure(figsize=(12, 12))
                plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
                plt.title('Z-score normalization applied')
                plt.show()

        elif z_score:
            img, mu, sigma = self.__normalize_image_zscore_per_image(img, verbosity)
            if verbosity:
                print('Z-score normalization applied based on image intensities, Mu=%s, sigma=%s' % (mu, sigma))
            if visualize:
                plt.figure(figsize=(12, 12))
                plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
                plt.title('Z-score normalization applied')
                plt.show()

        if reshape and to_shape:
            if to_shape.shape==3:
                img = self.__resize_3d_img(img)
                if verbosity:
                    print('Image reshaped: %s' % to_shape)
                if visualize:
                    plt.figure(figsize=(12, 12))
                    plt.imshow(img[int(len(img) / 2.), ...], cmap='bone')
                    plt.title('Reshaped image')
                    plt.show()

        return img





import pandas as pd
import numpy as np
import plotly.graph_objs as go
import plotly
import plotly.subplots
from statsmodels.stats import multitest
import os
import scipy as sp
import math
import scipy.stats
from operator import le, gt
from sklearn.metrics import auc
import plotly.io as pio
from pmtool.FeaturesSet import FeaturesSet

class AnalysisBox(FeaturesSet):

    '''This module is inherited from FeaturesSet class and allows for preliminary statistical analysis of the numerical features.'''

    def handle_nan(self, axis: int=1, how: str='any', mode: str='delete'):
        """Handle the missing values.

        Arguments:
            axis: Determines if patients (0) or variables (1) with the missing values have to be fixed.
            how: Determines if handling is needed when there is at least one missing value ('any') or all of them are missing ('all').
            mode: Determines the strategy: 'delete' will delete the variable/patient, 'fill' will fill a missing value with the imputation method.
        """
        if mode == 'delete':
            self._feature_dataframe.dropna(axis=axis, how=how, inplace=True)
            self._feature_outcome_dataframe.dropna(axis=axis, how=how, inplace=True)
            self._feature_column = list(self._feature_dataframe.columns)
            self._patient_name = list(self._feature_outcome_dataframe.index)
            if self._outcome_column in self._feature_column:
                self._feature_column.remove(self._outcome_column)
            self._outcome = self._feature_outcome_dataframe[self._outcome_column]
            self._class_label = pd.unique(np.array(list(self._outcome)))
            self._class_label.sort()
            data_balance = []
            for l in self._class_label:
                data_balance.append(np.sum(np.array(list(self._outcome)) == l)/len(self._outcome))
            print('Number of observations: {}\nClass labels: {}\nClasses balance: {}'.format(len(self._outcome),
                                                                                        self._class_label,
                                                                                        data_balance))
        if mode == 'fill':
            print('Not implemented yet')

        return None

    def handle_constant(self):
        """Drop the features with the constant values."""

        constant_features = self._feature_dataframe.columns[self._feature_dataframe.nunique() <= 1]
        self._feature_dataframe.drop(constant_features, axis=1, inplace=True)
        self._feature_outcome_dataframe.drop(constant_features, axis=1, inplace=True)
        self._feature_column = list(self._feature_dataframe.columns)
        if self._outcome_column in self._feature_column:
            self._feature_column.remove(self._outcome_column)
        self._outcome = self._feature_outcome_dataframe[self._outcome_column]
        self._class_label = pd.unique(np.array(list(self._outcome)))
        self._class_label.sort()
        data_balance = []
        for l in self._class_label:
            data_balance.append(np.sum(np.array(list(self._outcome)) == l) / len(self._outcome))
        print('Number of observations: {}\nClass labels: {}\nClasses balance: {}'.format(len(self._outcome),
                                                                                         self._class_label,
                                                                                         data_balance))

        return None

    def plot_distribution(self, features_to_plot: list=[], binary_classes_to_plot: list=[]):
        """Plot distribution of the feature values in classes into interactive .html report.

        Arguments:
            features_to_plot: List of specific features to be selected (otherwise selects all the numerical features).
            binary_classes_to_plot: List, containing 2 classes of interest, if the dataset is multi-class.
        """

        if len(self._outcome) > 0:
            if len(binary_classes_to_plot) == 2:
                if (binary_classes_to_plot[0] in self._class_label) & (binary_classes_to_plot[1] in self._class_label):
                    if not features_to_plot:
                        features_to_plot = self._feature_column
                    num_features = []
                    for feature in features_to_plot:
                        if self._feature_dataframe[feature].dtype != 'object':
                            num_features.append(feature)
                    cols = 4
                    rows = len(num_features) // 4 + 1
                    num_features_tuple = tuple(num_features)
                    fig = plotly.subplots.make_subplots(rows=rows, cols=cols, subplot_titles=num_features_tuple)
                    counter = 0
                    for feature in num_features:
                        c = counter % 4 + 1
                        r = counter // 4 + 1
                        fig.append_trace(go.Histogram(
                            x=self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                                  binary_classes_to_plot[0]][feature],
                            opacity=0.75,
                            name=str(binary_classes_to_plot[0]),
                            marker={'color': 'magenta'},
                            showlegend=(counter == 0)),
                            r, c)
                        fig.append_trace(go.Histogram(
                            x=self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                                  binary_classes_to_plot[1]][feature],
                            opacity=0.75,
                            name=str(binary_classes_to_plot[1]),
                            marker={'color': 'orange'},
                            showlegend=(counter == 0)),
                            r, c)
                        fig.update_xaxes(title_text='values', title_font={"size": 10}, title_standoff=5, row=r, col=c,
                                         showgrid=False, zeroline=False)
                        fig.update_yaxes(title_text='count', title_font={"size": 10}, title_standoff=0, row=r, col=c,
                                         showgrid=False, zeroline=False)
                        fig.layout.update(go.Layout(barmode='overlay'))
                        counter += 1
                    for i in fig['layout']['annotations']:
                        i['font'] = dict(size=10)
                    fig.update_layout(title_text='Features binary distribution in classes',
                                      height=rows * 250, width=1250)
                    plotly.offline.plot(fig,
                                        filename=os.path.splitext(self._feature_path)[0] + '_distr.html',
                                        config={'scrollZoom': True})
                else:
                    print('Wrong class label(s).')
            else:

                color_scheme = ['magenta', 'orange', 'cyan', 'yellow', 'lime',
                                'blue', 'red', 'green', 'darkviolet', 'saddlebrown']
                if not features_to_plot:
                    features_to_plot = self._feature_column
                num_features = []
                for feature in features_to_plot:
                    if self._feature_dataframe[feature].dtype != 'object':
                        num_features.append(feature)
                cols = 4
                rows = len(num_features) // 4 + 1
                num_features_tuple = tuple(num_features)
                fig = plotly.subplots.make_subplots(rows=rows, cols=cols, subplot_titles=num_features_tuple)
                counter = 0
                for feature in num_features:
                    c = counter % 4 + 1
                    r = counter // 4 + 1
                    counter_colors = 0
                    for cl in self._class_label:
                        fig.append_trace(go.Histogram(
                            x=self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                                  cl][feature],
                            opacity=0.5,
                            name=str(cl),
                            marker={'color': color_scheme[counter_colors]},
                            showlegend=(counter == 0)),
                            r, c)
                        counter_colors += 1

                    fig.update_xaxes(title_text='values', title_font={"size": 10}, title_standoff=5, row=r, col=c,
                                     showgrid=False, zeroline=False)
                    fig.update_yaxes(title_text='count', title_font={"size": 10}, title_standoff=0, row=r, col=c,
                                     showgrid=False, zeroline=False)
                    fig.layout.update(go.Layout(barmode='overlay'))
                    counter += 1
                for i in fig['layout']['annotations']:
                    i['font'] = dict(size=10)
                fig.update_layout(title_text='Features binary distribution in classes',
                                  height=rows * 250, width=1250)
                plotly.offline.plot(fig,
                                    filename=os.path.splitext(self._feature_path)[0] + '_distr.html',
                                    config={'scrollZoom': True})

        else:
            print('Outcome column should be presented')

        return None

    def plot_correlation_matrix(self, features_to_plot: list=[]):
        """Plot correlation (Spearman's) matrix for the features into interactive .html report.

        Arguments:
            features_to_plot: List of specific features to be selected (otherwise selects all the numerical features).
        """

        pio.renderers.default = 'iframe'

        if not features_to_plot:
            features_to_plot = self._feature_column

        num_features = []
        for feature in features_to_plot:
            if self._feature_dataframe[feature].dtype != 'object':
                num_features.append(feature)

        data = go.Heatmap(
            z=np.abs(np.array(self._feature_dataframe[num_features].corr(method='spearman'))),
            x=num_features,
            y=num_features,
            colorbar=dict(title='Spearman corr'),
            hovertemplate='feature_1: %{x}<br>feature_2: %{y}<br>r_Spearman: %{z}<extra></extra>'
        )

        layout = {"title": "Features correlation matrix"}
        fig = go.Figure(data=data, layout=layout)
        fig.update_xaxes(tickfont=dict(size=7))
        fig.update_yaxes(tickfont=dict(size=7))
        fig.update_layout(height=len(num_features) * 20 + 250, width=len(num_features) * 20 + 250)
        plotly.offline.plot(fig,
                            filename=os.path.splitext(self._feature_path)[0] + '_corr.html',
                            config={'scrollZoom': True})

        return None

    def __get_color(self, v, th):
        if abs(v) <= th:
            return 'orange'
        else:
            return 'purple'

    def __get_MW_p(self, ftrs, binary_classes_to_plot, p_threshold=0.05):
        p_MW = []
        df_0 = self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                   binary_classes_to_plot[0]]
        df_1 = self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                   binary_classes_to_plot[1]]
        for feature in ftrs:
            s, p = sp.stats.mannwhitneyu(df_0[feature].astype(float),
                                         df_1[feature].astype(float),
                                         alternative='two-sided')
            p_MW.append(p)
        reject, p_MW_corr, a_S_corr, a_B_corr = multitest.multipletests(p_MW,
                                                                        alpha=p_threshold,
                                                                        method='bonferroni')
        return p_MW_corr

    def plot_MW_p(self, features_to_plot: list=[], binary_classes_to_plot: list=[], p_threshold: float=0.05):
        """Plot two-sided Mann-Whitney U test p-values for comparison of features values means in 2 classes (with correction for multiple testing) into interactive .html report.

        Arguments:
            features_to_plot: List of specific features to be selected (otherwise selects all the numerical features).
            binary_classes_to_plot: List, containing 2 classes of interest, if the dataset is multi-class.
            p_threshold: Significance level.
        """

        if len(self._outcome) > 0:
            if len(self._class_label) == 2:
                binary_classes_to_plot = self._class_label
            else:
                if len(binary_classes_to_plot) != 2:
                    print('Only binary class labels are supported.')
                    return
                elif not ((binary_classes_to_plot[0] in self._class_label) & (binary_classes_to_plot[1] in self._class_label)):
                    print('Wrong class label(s).')
                    return

            if not features_to_plot:
                features_to_plot = self._feature_column

            num_features = []
            for feature in features_to_plot:
                if self._feature_dataframe[feature].dtype != 'object':
                    num_features.append(feature)
            p_MW_corr = self.__get_MW_p(ftrs=num_features,
                                        binary_classes_to_plot=binary_classes_to_plot,
                                        p_threshold=p_threshold)

            colors = [self.__get_color(v, p_threshold) for v in p_MW_corr]
            shapes = [{'type': 'line',
                        'xref': 'x',
                        'yref': 'y',
                        'x0': p_threshold,
                        'y0': 0,
                        'x1': p_threshold,
                        'y1': len(num_features)}]
            annotations = [
                go.layout.Annotation(
                    x=math.log10(p_threshold),
                    y=len(num_features),
                    text="alpha="+str(p_threshold),
                    showarrow=True, arrowhead=7, ax=100, ay=0
                )
            ]
            layout = plotly.graph_objs.Layout(shapes=shapes)

            fig = go.Figure(
                [go.Bar(x=list(p_MW_corr), y=num_features, marker={'color': colors}, orientation='h')],
                layout=layout)
            fig.update_yaxes(tickfont=dict(size=7))
            fig.update_xaxes(tickfont=dict(size=7))
            fig.update_layout(title_text='The p-values for Mann-Whitney test (Bonferroni corrected)',
                                height=len(num_features) * 20+250, width=750,
                                xaxis_type="log", annotations=annotations,
                                xaxis={"mirror": "allticks", 'side': 'top', 'dtick': 1, 'showgrid': True}
                                )
            plotly.offline.plot(fig,
                                filename=os.path.splitext(self._feature_path)[0] + '_MW.html',
                                config={'scrollZoom': True})
        else:
            print('Outcome column should be presented')
        return None

    def __get_univar_fprs_tprs(self, ftr, binary_classes_to_plot):

        mean_0 = self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                     binary_classes_to_plot[0]][ftr].mean()
        mean_1 = self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                     binary_classes_to_plot[1]][ftr].mean()
        feature_min = np.min(np.array(self._feature_outcome_dataframe[ftr]))
        feature_max = np.max(np.array(self._feature_outcome_dataframe[ftr]))
        step = (feature_max - feature_min) / 100
        x = np.array(self._feature_outcome_dataframe[ftr])
        n_0 = len(self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                      binary_classes_to_plot[0]])
        n_1 = len(self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                      binary_classes_to_plot[1]])
        th = []
        tprs = []
        fprs = []
        op = gt
        if mean_0 > mean_1:
            op = le

        for i in range(0, 102):
            t = feature_min + i * step
            th.append(t)
            y_pred = op(x, t).astype(int)
            tp = np.sum(y_pred[np.where(self._feature_outcome_dataframe[self._outcome_column] == binary_classes_to_plot[1])])
            fp = np.sum(y_pred[np.where(self._feature_outcome_dataframe[self._outcome_column] == binary_classes_to_plot[0])])
            tpr = tp / n_1
            fpr = fp / n_0
            tprs.append(tpr)
            fprs.append(fpr)

        return fprs, tprs

    def plot_univariate_roc(self, features_to_plot: list=[], binary_classes_to_plot: list=[], auc_threshold: float=0.75):
        """Plot univariate ROC curves (with AUC calculation) for threshold binary classifier, based of each feature separately into interactive .html report.

        Arguments:
            features_to_plot: List of specific features to be selected (otherwise selects all the numerical features).
            binary_classes_to_plot: List, containing 2 classes of interest in case of multi-class data.
            auc_threshold: Threshold value for ROC AUC to be highlighted.
        """

        if len(self._outcome) > 0:
            if len(self._class_label) == 2:
                binary_classes_to_plot = self._class_label
            else:
                if len(binary_classes_to_plot) != 2:
                    print('Only binary class labels are supported.')
                    return
                elif not ((binary_classes_to_plot[0] in self._class_label) & (
                        binary_classes_to_plot[1] in self._class_label)):
                    print('Wrong class label(s).')
                    return

            if not features_to_plot:
                features_to_plot = self._feature_column

            num_features = []
            for feature in features_to_plot:
                if self._feature_outcome_dataframe[feature].dtype != 'object':
                    num_features.append(feature)

            cols = 4
            rows = len(num_features) // 4 + 1
            num_features_tuple = tuple(num_features)

            fig = plotly.subplots.make_subplots(rows=rows, cols=cols, subplot_titles=num_features_tuple)
            counter = 0

            for feature in num_features:
                c = counter % 4 + 1
                r = counter // 4 + 1
                fprs, tprs = self.__get_univar_fprs_tprs(ftr=feature, binary_classes_to_plot=binary_classes_to_plot)
                univar_auc = auc(fprs, tprs)
                fig.append_trace(go.Scatter(x=fprs, y=tprs,
                                            name='ROC',
                                            marker={'color': self.__get_color(univar_auc, auc_threshold)}),
                                    r, c)
                fig.append_trace(go.Scatter(x=[0, 1], y=[0, 1],
                                            name='Chance', marker={'color': 'grey'},
                                            mode='lines+markers+text',
                                            text=[' ROC AUC=%0.2f' % univar_auc, ''], textposition='middle right'),
                                    r, c)

                fig.update_xaxes(title_text='FPR', title_font={"size": 10}, title_standoff=5, row=r, col=c,
                                 showgrid=False, zeroline=False)
                fig.update_yaxes(title_text='TPR', title_font={"size": 10}, title_standoff=0, row=r, col=c,
                                 showgrid=False, zeroline=False)
                counter += 1

            for i in fig['layout']['annotations']:
                i['font'] = dict(size=10)

            fig.update_layout(title_text='Features univariate ROC-curves:' + str(binary_classes_to_plot),
                              height=rows * 250, width=1250, showlegend=False)
            plotly.offline.plot(fig,
                                filename=os.path.splitext(self._feature_path)[0] + '_roc-univar.html',
                                config={'scrollZoom': True})
        else:
            print('Outcome column should be presented')

        return None

    def calculate_basic_stats(self, volume_feature: str=''):
        """Calculate basic statistical scores (such as: number of missing values, mean, std, min, max, Mann-Whitney test p-values for binary classes, univariate ROC AUC for binary classes, Spearman's correlation with volume if volumetric feature name is sent to function) for each feature and save it to .csv file.

        Arguments:
            volume_feature: Name of the feature, which is considered as volume.
        """

        num_features = []
        for feature in self._feature_column:
            if self._feature_dataframe[feature].dtype != 'object':
                num_features.append(feature)

        frame = {'NaN': self._feature_dataframe[num_features].isnull().sum(axis=0),
                 'Mean': self._feature_dataframe[num_features].mean(axis=0),
                 'Std': self._feature_dataframe[num_features].std(axis=0),
                 'Min': self._feature_dataframe[num_features].min(axis=0),
                 'Max': self._feature_dataframe[num_features].max(axis=0)}
        stats_dataframe = pd.DataFrame(frame)

        if len(self._outcome) > 0:
            if len(self._class_label) == 2:
                p_MW_corr = self.__get_MW_p(ftrs=num_features, binary_classes_to_plot=self._class_label)
                univar_auc = []
                for feature in num_features:
                    fprs, tprs = self.__get_univar_fprs_tprs(ftr=feature, binary_classes_to_plot=self._class_label)
                    univar_auc.append(auc(fprs, tprs))
                stats_dataframe_ext = pd.DataFrame({'p_MW_corrected': p_MW_corr,
                                                    'univar_auc': univar_auc},
                                                   index=num_features)

                stats_dataframe = pd.concat([stats_dataframe, stats_dataframe_ext], axis=1)

        if volume_feature:
            vol_corr = []
            for feature in num_features:
                vol_corr.append(sp.stats.spearmanr(self._feature_dataframe[feature],
                                                   self._feature_dataframe[volume_feature])[0])
            stats_dataframe = pd.concat([stats_dataframe,
                                         pd.DataFrame({'volume_corr': vol_corr}, index=num_features)],
                                        axis=1)

        stats_dataframe.to_excel(os.path.splitext(self._feature_path)[0] + '_basic_stats.xlsx')

        return None

    def __get_univar_prec_rec(self, ftr):

        mean_0 = self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                     self._class_label[0]][ftr].mean()
        mean_1 = self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                     self._class_label[1]][ftr].mean()
        feature_min = np.min(np.array(self._feature_outcome_dataframe[ftr]))
        feature_max = np.max(np.array(self._feature_outcome_dataframe[ftr]))
        step = (feature_max - feature_min) / 100
        x = np.array(self._feature_outcome_dataframe[ftr])
        n_1 = len(self._feature_outcome_dataframe.loc[self._feature_outcome_dataframe[self._outcome_column] ==
                                                      self._class_label[1]])
        th = []
        precs = []
        recs = []
        op = gt
        if mean_0 > mean_1:
            op = le

        for i in range(0, 102):
            t = feature_min + i * step
            th.append(t)
            y_pred = op(x, t).astype(int)
            if np.sum(y_pred>0):
                tp = np.sum(y_pred[np.where(self._feature_outcome_dataframe[self._outcome_column] == self._class_label[1])])
                prec = tp / np.sum(y_pred)
                rec = tp / n_1
                precs.append(prec)
                recs.append(rec)

        return precs, recs

    def volume_analysis(self, volume_feature: str='', auc_threshold: float=0.75, features_to_plot: list=[], corr_threshold: float=0.75):
        '''Calculate features correlation (Spearman’s) with volume and plot volume-based precision-recall curve.

        Arguments:
            volume_feature: Name of the feature, which is considered as volume.
            auc_threshold: Threshold value for area under precision-recall curve to be highlighted.
            features_to_plot: Specific features to be selected (otherwise selects all the numerical features)
            corr_threshold: Threshold value for absolute value for Spearman’s correlation coefficient to be considered as ‘strong correlation’.
        '''
        if volume_feature:
            if volume_feature in self._feature_column:
                if len(self._outcome) > 0:
                    if len(self._class_label) == 2:

                        precs, recs = self.__get_univar_prec_rec(ftr=volume_feature)
                        univar_auc = auc(recs, precs)
                        fig = go.Figure()
                        fig.add_trace(go.Scatter(x=recs, y=precs,
                                                 name='PRC',
                                                 marker={'color': self.__get_color(univar_auc, auc_threshold)}))
                        fig.add_trace(go.Scatter(x=[0], y=[0],
                                                 name='AUC', marker={'color': 'grey'},
                                                 mode='text',
                                                 text=['AUC=%0.2f' % univar_auc, ''], textposition='top right'))
                        fig.update_xaxes(title_text='recall', title_font={"size": 10}, title_standoff=5,
                                         showgrid=True, zeroline=False, range=[0, 1])
                        fig.update_yaxes(title_text='precision', title_font={"size": 10}, title_standoff=5,
                                         showgrid=True, zeroline=False, range=[0, 1])

                        fig.update_layout(title_text='Volume precision-recall curve', showlegend=False, width=500,
                                          height=500)
                        plotly.offline.plot(fig,
                                            filename=os.path.splitext(self._feature_path)[0] + '_volume_PRC.html',
                                            config={'scrollZoom': True})

                if not features_to_plot:
                    features_to_plot = self._feature_column
                num_features = []
                corr_vol = []
                for feature in features_to_plot:
                    if self._feature_outcome_dataframe[feature].dtype != 'object':
                        num_features.append(feature)
                        corr_vol.append(sp.stats.spearmanr(self._feature_dataframe[feature],
                                                           self._feature_dataframe[volume_feature])[0])
                colors = [self.__get_color(v, corr_threshold) for v in corr_vol]
                shapes = [{'type': 'line',
                           'xref': 'x',
                           'yref': 'y',
                           'x0': corr_threshold,
                           'y0': 0,
                           'x1': corr_threshold,
                           'y1': len(num_features)}]
                annotations = [
                    go.layout.Annotation(
                        x=corr_threshold,
                        y=len(num_features),
                        text="r_S=" + str(corr_threshold),
                        showarrow=True, arrowhead=7, ax=100, ay=0
                    )
                ]
                layout = plotly.graph_objs.Layout(shapes=shapes)
                fig = go.Figure(layout=layout)
                fig.add_trace(go.Bar(x=list(np.abs(corr_vol)), y=num_features, marker={'color': colors},
                                     orientation='h', name='Spearman correlation'))
                fig.update_yaxes(tickfont=dict(size=7))
                fig.update_xaxes(tickfont=dict(size=7))
                fig.update_layout(annotations=annotations,
                                  xaxis={"mirror": "allticks", 'side': 'top', 'dtick': 1, 'showgrid': True},
                                  title_text='Volume Spearman correlation', showlegend=False, width=750,
                                  height=len(num_features) * 20+250)
                plotly.offline.plot(fig, filename=os.path.splitext(self._feature_path)[0] + '_volume_corr.html',
                                    config={'scrollZoom': True})

        return None




from distutils.core import setup
setup(
  name = 'precision-medicine-toolbox',         # How you named your package folder (MyLib)
  packages = ['pmtool'],   # Chose the same as "name"
  version = '0.2',      # Start with a small number and increase it with every change you make
  license='bsd-3-clause',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Precision medicine tool-box for medical imaging research',   # Give a short description about your library
  author = 'sergey primakov & lisa lavrova',                   # Type in your name
  author_email = 'primakov@bk.ru',      # Type in your E-Mail
  url = 'https://github.com/primakov/precision-medicine-toolbox',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/primakov/precision-medicine-toolbox/archive/refs/tags/0.0.tar.gz',    # I explain this later on
  keywords = ['medical imaging research', 'dicom', 'radiomics', 'statistical analysis', 'features'],   # Keywords that define your package best
  install_requires=[
    'SimpleITK',
    'PyWavelets',
    'pykwalify',
    'six',
    'tqdm',
    'pydicom',
    'pandas',
    'pyradiomics',
    'scikit-image',
    'ipywidgets',
    'matplotlib',
    'Pillow',
    'scikit-learn',
    'scipy',
    'plotly',
    'mkdocstrings',
    'statsmodels',
    'opencv-python'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)